#include "mynhmlg.h"

/* Procov.
 * The ASRV and Covarionlike_node recursive functions.
 * Covarion uses most of procov's total CPU time.
 * This file contains various optimized versions of Covarionlike_node, and a
 * wrapper to select one based on getenv("PROCOV_CLN_NEW");
 */

static void *staticS1=NULL, *staticS2=NULL;
static size_t Ssize=0;
/* keep arrays allocated to reduce alloc/free/memset overhead
 * Make S1/S2 plenty big: root->nbpattern == lgseq, = biggest we could possibly need
 * There can be large left and right sub-trees, regardless of how it is in the root. */
static inline void S12setup(noeud root)
{
  size_t minbytes = sizeof(double) * 20 * gamma_nbcl1 * root->nbpattern;
  if(Ssize < minbytes){
    Ssize = minbytes;
    staticS1 = realloc(staticS1, minbytes);
    staticS2 = realloc(staticS2, minbytes);
    if(!staticS1 || !staticS2){ perror("allocating S1/S2 matrices"); exit(1); }
  }
}

#define X_NEGATIVE_CHECK(x) \
        if(!isgreaterequal(x,0.)){ /* isgreaterequal() returns false on NaN */ \
          if(!isgreaterequal(x,ERROR_THRESHOLD)){\
            fprintf(stderr, "procov: error in %s: nd:%s, i:%d, cl:%d, j:%d , nd->x[i][cl][j]:%.12g\n", __func__, nd->nom, i, cl, j, x);\
            exit(3); /* abort(); */\
          }else x=0.;\
        }

// This function uses ~70% of total CPU time with the cpn60 data set (17 species, 1000 sites), tuffley.opt
// (more like 90% of the time not spent in BLAS libraries.)
static void Covarionlike_node_old(noeud nd)
{
  int i, j, k, cl, cl2;
  double s1, s2;
  if(nd->v1==NULL) return;
  Covarionlike_node_old(nd->v1);
  Covarionlike_node_old(nd->v2);

  int nbcl = gamma_nbcl1;
  for(i=0;i<nd->nbpattern;i++){
    int ps1=nd->patternson1[i], ps2=nd->patternson2[i];
    for(cl=0;cl<nbcl;cl++) for(j=0;j<20;j++){
      s1=s2=0.;
      for(cl2=0;cl2<nbcl;cl2++) for(k=0;k<20;k++){
	s1 += nd->Pind(p1,nbcl, cl,j, cl2,k) * nd->v1->x[ps1][cl2][k];
	s2 += nd->Pind(p2,nbcl, cl,j, cl2,k) * nd->v2->x[ps2][cl2][k];
      }
      nd->x[i][cl][j] = s1*s2;
      X_NEGATIVE_CHECK(nd->x[i][cl][j]);
    }
  }
}

#ifdef USE_BLAS
/* Do all the x vectors at once with a big matrix multiply.
 * An optimized dgemm using SIMD floating point (SSE2 mulpd) can beat anything gcc
 * can generate, because I think SIMD gets twice the floating point throughput. */
static void Covarionlike_node_new3(noeud nd)
{
  int i, j, cl;
  if(nd->v1==NULL) return;
  Covarionlike_node_new3(nd->v1);
  Covarionlike_node_new3(nd->v2);

  int nbcl = gamma_nbcl1;
  int ps1size=nd->ps1size, ps2size=nd->ps2size;
  double (*S1)[nbcl][20] = staticS1, (*S2)[nbcl][20] = staticS2;
  // ->x and S1/2, looked at as arrays of column vectors, are stored in column major order
  // p1/p2 are square, so tell blas column major, and transpose p1/2
#ifdef HAVE_CBLAS
  cblas_dgemm(CblasColMajor, CblasTrans, CblasNoTrans, 20*nbcl, ps1size, 20*nbcl,
	      1.0, nd->p1, 20*nbcl, &nd->v1->x[0][0][0], 20*nbcl, 0.0, (double *)S1, 20*nbcl);
  cblas_dgemm(CblasColMajor, CblasTrans, CblasNoTrans, 20*nbcl, ps2size, 20*nbcl,
	      1.0, nd->p2, 20*nbcl, &nd->v2->x[0][0][0], 20*nbcl, 0.0, (double *)S2, 20*nbcl);
#else
  dgemm(				     'T', 'N', 20*nbcl, ps1size, 20*nbcl,
	      1.0, nd->p1, 20*nbcl, &nd->v1->x[0][0][0], 20*nbcl, 0.0, (double *)S1, 20*nbcl);
  dgemm(				     'T', 'N', 20*nbcl, ps2size, 20*nbcl,
	      1.0, nd->p2, 20*nbcl, &nd->v2->x[0][0][0], 20*nbcl, 0.0, (double *)S2, 20*nbcl);
#endif

  for(i=0;i<nd->nbpattern;i++){
    int ps1=nd->patternson1[i], ps2=nd->patternson2[i];
    for(cl=0 ; cl<nbcl ; cl++) for(j=0 ; j<20 ; j++){
      nd->x[i][cl][j] = S1[ps1][cl][j] * S2[ps2][cl][j];
      X_NEGATIVE_CHECK(nd->x[i][cl][j]);
    }
  }
}

#endif // USE_BLAS


// only do up to ps1/2max, not the whole nbpattern
// this is like _new3's dgemm over the whole thing.
static void Covarionlike_node_new6(noeud nd)
{
  int i, j, cl, k, cl2;
  if(nd->v1==NULL) return;
  Covarionlike_node_new6(nd->v1);
  Covarionlike_node_new6(nd->v2);

  int nbcl = gamma_nbcl1;
  int ps1size=nd->ps1size, ps2size=nd->ps2size, ps1, ps2;
  double (*S1)[nbcl][20] = staticS1, (*S2)[nbcl][20] = staticS2;
  for(ps1=0 ; ps1<ps1size ; ps1++) for(cl=0;cl<nbcl;cl++) for(j=0;j<20;j++){
    double s1 = 0;
    for(cl2=0;cl2<nbcl;cl2++) for(k=0;k<20;k++)
      s1 += Pind(nd->p1,nbcl, cl,j, cl2,k) * nd->v1->x[ps1][cl2][k];
    S1[ps1][cl][j] = s1;

  }
  for(ps2=0 ; ps2<ps2size ; ps2++) for(cl=0;cl<nbcl;cl++) for(j=0;j<20;j++){
    double s2 = 0;
    for(cl2=0;cl2<nbcl;cl2++) for(k=0;k<20;k++)
      s2 += Pind(nd->p2,nbcl, cl,j, cl2,k) * nd->v2->x[ps2][cl2][k];
    S2[ps2][cl][j] = s2;
  }

  for(i=0;i<nd->nbpattern;i++){
    int ps1=nd->patternson1[i], ps2=nd->patternson2[i];
    for(cl=0 ; cl<nbcl ; cl++) for(j=0 ; j<20 ; j++){
      nd->x[i][cl][j] = S1[ps1][cl][j] * S2[ps2][cl][j];
      X_NEGATIVE_CHECK(nd->x[i][cl][j]);
    }
  }
}


// wrapper function to dispatch to different optimized versions
void Covarionlike_node(noeud root)
{
  //  char c, *p = getenv("PROCOV_CLN_NEW");

  S12setup(root);
  /*
  if(!p) c='3'; else c=*p;	// level3 BLAS is fastest
  switch(c){
  case 'o': Covarionlike_node_old(root); break;
#ifdef USE_BLAS
  case '3': Covarionlike_node_new3(root); break;
#endif
  case '6': Covarionlike_node_new6(root); break;

  default:
    Covarionlike_node_new6(root); // best non-BLAS version
  }
  */

  /* Covarionlike_node_old(root); */ // old version, not used

#ifdef USE_BLAS
  Covarionlike_node_new3(root);     // new3 BLAS is fastest
#else 
  Covarionlike_node_new6(root); // best non-BLAS version

#endif


}


void ASRVlike_node(noeud nd)
{
  int i, j, k, cl;
  if(nd->v1==NULL) return;
  else if(!nd->v3) S12setup(nd);
  ASRVlike_node(nd->v1);
  ASRVlike_node(nd->v2);

  int ps1size=nd->ps1size, ps2size=nd->ps2size, ps1, ps2;
  double (*S1)[gamma_nbcl][20] = staticS1, (*S2)[gamma_nbcl][20] = staticS2;
  // This doesn't work as a matrix multiply because there's no looping over cl2
  for(ps1=0 ; ps1<ps1size ; ps1++) for(cl=0;cl<gamma_nbcl;cl++) for(j=0;j<20;j++){
    double s1 = 0;		// let gcc keep this var in a register
    for(k=0;k<20;k++)
      s1 += Pind(nd->p1,gamma_nbcl1, cl,j, cl,k) * nd->v1->x[ps1][cl][k];
    S1[ps1][cl][j] = s1;
  }
  for(ps2=0 ; ps2<ps2size ; ps2++) for(cl=0;cl<gamma_nbcl;cl++) for(j=0;j<20;j++){
    double s2 = 0;
    for(k=0;k<20;k++)
      s2 += Pind(nd->p2,gamma_nbcl1, cl,j, cl,k) * nd->v2->x[ps2][cl][k];
    S2[ps2][cl][j] = s2;
  }

  for(i=0;i<nd->nbpattern;i++){
    int ps1=nd->patternson1[i], ps2=nd->patternson2[i];
    for(cl=0;cl<gamma_nbcl;cl++) for(j=0;j<20;j++){ // note _not_ nbcl1
      nd->x[i][cl][j] = S1[ps1][cl][j] * S2[ps2][cl][j];
      X_NEGATIVE_CHECK(nd->x[i][cl][j]);
    }
  }
}
